//==============================================================================
// LittleCMS Test/Hacks
//==============================================================================

//#include <QCoreApplication>
#include <lcms2.h>
#include <stdio.h>

#include <QImage>

static void CPUCycle (quint64& res)
{
  #if defined _M_IX86 // MICROSOFT COMPILER && x86 PROCESSOR.
    __asm
    {
      MOV ECX, res
      RDTSC // Supported by Pentium Class Processor
      MOV [ECX+0], EAX
      MOV [ECX+4], EDX
    }
  #else
    {
      unsigned long long int tmp;
     __asm__ volatile (".byte 0x0f, 0x31" : "=A" (tmp));
     res = tmp;
    }
  #endif
}
//==============================================================================
static quint64 RunOnce(cmsHTRANSFORM t, uchar * bits, int sz, int num)
{
  quint64 Ta, Tb;
  CPUCycle(Ta);
    cmsDoTransform(t, bits, bits, sz);
  CPUCycle(Tb);
  quint64 delta = (Tb-Ta);
  printf(QString("(test %1 lasts %2 KCycles).\n").arg(num).arg((int)(delta/1000)).toLatin1().constData());
  return delta;
}
//==============================================================================
static void RunSeveral (const QString& title, const QImage& img, cmsHTRANSFORM& tr)
{
  quint64 totalKCycles = 0;
  int max    = 4;
  int nbPxls = img.width()*img.height();
  QImage imgTest(img);
  printf(QString("**** Start TEST : %2 ****\n").arg(title).toLatin1().constData());
  uchar *bits = imgTest.bits();
  for (int n = 0 ; n<max ; n++)
  {
    totalKCycles += RunOnce(tr, bits, nbPxls, n);
    // If you want to check the result, turn max to 1 and uncomment the line below:
    // Note that only the first pass is correct (n==0). All other pass will convert an 
    // already converted image. But this is not a problem to the bench point of view...
    // other point, the result png image has no color profile attached as Qt doesn't manage this
    // (it will be opened as sRGB unless you tell to your software to assign to the sreen profile)
    // if (n==0) imgTest.save( QString("res/result_%1.png").arg(title), "PNG");
  }
  int   statsKCycle     = totalKCycles / 1000 / max;
  qreal statCyclePerPxl = (qreal)totalKCycles / nbPxls / max;
  printf(QString("Average Test lasts %2 KCycles.\n").arg(statsKCycle).toLatin1().constData());
  printf(QString("Average CPU Cycle per pixel = %1.\n").arg(statCyclePerPxl,0,'f',2).toLatin1().constData());
  printf("-------------------------------------------\n");
}
//==============================================================================
static void SetAlphaChannel( QImage& img, const QImage& mask )
{
  if (img.size() == mask.size() && img.format() == QImage::Format_ARGB32 && mask.format() == QImage::Format_Indexed8)
  {
    for (int y = 0 ; y < img.height() ; y++)
    {
      QRgb* imgBit         = (QRgb*) img.scanLine (y);
      const uchar* maskBit =         mask.scanLine(y);
      for (int x = 0 ; x < img.width() ; x++) imgBit[x] = (imgBit[x] & RGB_MASK) | (maskBit[x] << 24);
    }
  }
  else printf("Something's wrong with the alpha mask...\n");
}
//==============================================================================
int main( int argc, char **argv )
{
  printf("LittleCMS Test/Hacks & simple benchmarking...\n");
  
  // We will bench a conversion from ECI RGB to a Screen color space, in RGB(A) modes :
  // And for softProofing, we use the FOGRA profile
  QImage test0("res/benchImg.jpg");
  QImage test1 = test0; test1 = test1.convertToFormat( QImage::Format_ARGB32 );
  QImage maskWorst ("res/maskWorst.png" ); maskWorst  = maskWorst .convertToFormat( QImage::Format_Indexed8 );
  QImage maskSprite("res/maskSprite.png"); maskSprite = maskSprite.convertToFormat( QImage::Format_Indexed8 );
  QImage test2 = test1; SetAlphaChannel(test2, maskWorst ); test2 = test2.convertToFormat( QImage::Format_ARGB32_Premultiplied );
  QImage test3 = test1; SetAlphaChannel(test3, maskSprite); test3 = test3.convertToFormat( QImage::Format_ARGB32_Premultiplied );

  cmsHPROFILE    hdl_ECI    = cmsOpenProfileFromFile("res/ECIRGB_v2.icc","r");
  cmsHPROFILE    hdl_Screen = cmsOpenProfileFromFile("res/Samsung_226BW_A.icm","r");
  cmsHPROFILE    hdl_cmyk   = cmsOpenProfileFromFile("res/CoatedFOGRA39.icc","r");

  cmsHTRANSFORM  transform0 = cmsCreateTransform        (hdl_ECI, TYPE_RGBA_8,        hdl_Screen, TYPE_RGBA_8,        INTENT_PERCEPTUAL, 0); // (invalid & not optimized)
  cmsHTRANSFORM  transform1 = cmsCreateTransform        (hdl_ECI, TYPE_BGRA_8,        hdl_Screen, TYPE_BGRA_8,        INTENT_PERCEPTUAL, 0);
  cmsHTRANSFORM  transform2 = cmsCreateTransform        (hdl_ECI, TYPE_BGRA_8_PREMUL, hdl_Screen, TYPE_BGRA_8_PREMUL, INTENT_PERCEPTUAL, 0);

  cmsHTRANSFORM  transform3 = cmsCreateProofingTransform(hdl_ECI, TYPE_RGBA_8,        hdl_Screen, TYPE_RGBA_8,        hdl_cmyk, INTENT_PERCEPTUAL, INTENT_RELATIVE_COLORIMETRIC, cmsFLAGS_SOFTPROOFING); // (invalid & not optimized)
  cmsHTRANSFORM  transform4 = cmsCreateProofingTransform(hdl_ECI, TYPE_BGRA_8,        hdl_Screen, TYPE_BGRA_8,        hdl_cmyk, INTENT_PERCEPTUAL, INTENT_RELATIVE_COLORIMETRIC, cmsFLAGS_SOFTPROOFING);
  cmsHTRANSFORM  transform5 = cmsCreateProofingTransform(hdl_ECI, TYPE_BGRA_8_PREMUL, hdl_Screen, TYPE_BGRA_8_PREMUL, hdl_cmyk, INTENT_PERCEPTUAL, INTENT_RELATIVE_COLORIMETRIC, cmsFLAGS_SOFTPROOFING);

  if ((hdl_Screen==0)||(hdl_ECI==0)||(hdl_cmyk==0))       { printf("Something's wrong with cms profiles...\n");    exit(0); }
  if (test0.isNull()||test1.isNull()||test2.isNull())     { printf("Something's wrong with images...\n");          exit(0); } 
  if ((transform0==0)||(transform1==0)||(transform2==0)||
      (transform3==0)||(transform4==0)||(transform5==0))  { printf("Something's wrong with cms HTransforms...\n"); exit(0); }
  cmsCloseProfile(hdl_cmyk);
  cmsCloseProfile(hdl_Screen);
  cmsCloseProfile(hdl_ECI);
  printf("Init OK...\n\n");

  // Note that the very next test is only accurate regarding CPU Cycle. The image itself is all
  // wrong because RGB has been swapped. But this test is interesting to compare with optimized code
  RunSeveral ("LittleCMS_AlphaIgnored_NoProofing",                   test0, transform0);

  RunSeveral ("RGBAEngine_AlphaIgnored_NoProofing",                  test0, transform1);
  RunSeveral ("RGBAEngine_PreMulAlpha_NoProofing_ImageWithoutAlpha", test1, transform2);
  RunSeveral ("RGBAEngine_PreMulAlpha_NoProofing_ImageWorstCase"   , test2, transform2);
  RunSeveral ("RGBAEngine_PreMulAlpha_NoProofing_ImageSpriteCase"  , test3, transform2);

  // Note that the very next test is only accurate regarding CPU Cycle. The image itself is all
  // wrong because RGB has been swapped. But this test is interesting to compare with optimized code
  RunSeveral ("LittleCMS_AlphaIgnored_SoftProofing",                   test0, transform3);

  RunSeveral ("RGBAEngine_AlphaIgnored_SoftProofing",                  test0, transform4);
  RunSeveral ("RGBAEngine_PreMulAlpha_SoftProofing_ImageWithoutAlpha", test1, transform5);
  RunSeveral ("RGBAEngine_PreMulAlpha_SoftProofing_ImageWorstCase"   , test2, transform5);
  RunSeveral ("RGBAEngine_PreMulAlpha_SoftProofing_ImageSpriteCase"  , test3, transform5);

  cmsDeleteTransform(transform0);
  cmsDeleteTransform(transform1);
  cmsDeleteTransform(transform2);
  cmsDeleteTransform(transform3);
  cmsDeleteTransform(transform4);
  cmsDeleteTransform(transform5);

  /*
  // Wondering why Marti wrote FROM_16_TO_8 instead of basic round...
  // Difference with a basic unsigned cast is not really linear.
  // Not been better to integrate this "cast" into the Shaper2[RGB] 
  // Lookup table (and use a Shaper2X specific for 8 bits)
  printf("Conversion is different for following values :");
  for (uint i = 0 ; i<65536 ; i++)
  {
    uchar  c1 = FROM_16_TO_8(i);
    uchar  c2 = i>>8;
    if (c1 != c2) printf("value = %X ; FROM_16_TO_8 => %X ; basic => %X\n", i, c1,c2);
  }
  */

  printf("Work's done...\n");
  char ch = getchar();
  return 0;
}
//==============================================================================
